/* Copyright (c) 1997-2010
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Darmstadt, Germany)
   http://www.polymake.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
--------------------------------------------------------------------------------
   $Project: polymake $$Id: graph_compare.h 10050 2011-03-01 15:46:48Z paffenho $
*/

#ifndef POLYMAKE_GRAPH_COMPARE_H
#define POLYMAKE_GRAPH_COMPARE_H

#include "polymake/GenericGraph.h"
#include "polymake/GenericIncidenceMatrix.h"
#include "polymake/Array.h"
#include "polymake/Map.h"
#include "polymake/vector"
#include "polymake/list"

namespace polymake { namespace graph {

class BlissGraph {
   struct impl;
   impl *p_impl;

   int n_autom;
   std::list< Array<unsigned int> > autom;

   static impl *alloc_impl(int n_nodes, bool is_directed);
   void add_edge(unsigned int from, unsigned int to);
   void color(unsigned int at, unsigned int color);
   void color(unsigned int from, unsigned int to, unsigned int color);
   void partition(unsigned int at);
   void finalize(bool gather_automorphisms);

   template <typename Matrix>
   void fill(const GenericIncidenceMatrix<Matrix>& M)
   {
      for (typename Entire< Rows< Matrix > >::const_iterator r=entire(rows(M));  !r.at_end();  ++r)
         for (typename Entire< typename Matrix::row_type >::const_iterator c=entire(*r);  !c.at_end();  ++c)
            add_edge(r.index(), *c);
   }

   template <typename Matrix, typename Iterator>
   void fill_renumbered(const GenericIncidenceMatrix<Matrix>& M, int d, Iterator iit)
   {
      std::vector<int> renumber(d);
      for (int i=0; !iit.at_end(); ++iit, ++i)
         renumber[iit.index()]=i;
      for (typename Entire< Rows< Matrix > >::const_iterator r=entire(rows(M));  !r.at_end();  ++r)
         for (typename Entire< typename Matrix::row_type >::const_iterator c=entire(*r);  !c.at_end();  ++c)
            add_edge(renumber[r.index()], renumber[*c]);
   }

   template <typename Graph>
   void fill(const GenericGraph<Graph>& G)
   {
      if (G.top().has_gaps())
         fill_renumbered(adjacency_matrix(G), G.top().dim(), entire(nodes(G)));
      else
         fill(adjacency_matrix(G));
   }

public:
   BlissGraph() : p_impl(0) {}

   template <typename Graph>
   explicit BlissGraph(const GenericGraph<Graph>& G, bool gather_automorphisms=false)
      : p_impl(alloc_impl(G.nodes(), G.is_directed))
   {
      fill(G);
      finalize(gather_automorphisms);
   }

   // non-symmetrical incidence matrix
   template <typename Matrix>
   explicit BlissGraph(const GenericIncidenceMatrix<Matrix>& M,
                       typename pm::disable_if<bool, Matrix::is_symmetric>::type gather_automorphisms=false)
      : p_impl(alloc_impl(M.rows()+M.cols(), false))
   {
      unsigned int rnode=M.cols();
      partition(rnode);
      for (typename Entire< Rows< Matrix > >::const_iterator r=entire(rows(M));  !r.at_end();  ++r, ++rnode)
         for (typename Entire< typename Matrix::row_type >::const_iterator c=entire(*r);  !c.at_end();  ++c) {
            add_edge(rnode, *c);
            add_edge(*c, rnode);
         }
      finalize(gather_automorphisms);
   }

   // symmetrical incidence matrix
   template <typename Matrix>
   explicit BlissGraph(const GenericIncidenceMatrix<Matrix>& M,
                       typename pm::enable_if<bool, Matrix::is_symmetric>::type gather_automorphisms=false)
      : p_impl(alloc_impl(M.rows(), true))
   {
      // there can be ones on the diagonal, which correspond loops in the graph;
      // bliss requires the graph to be declared as directed in this case
      fill(M);
      finalize(gather_automorphisms);
   }

   ~BlissGraph();

   bool operator== (const BlissGraph& g2) const;
   bool operator!= (const BlissGraph& g2) const { return !operator==(g2); }

private:
   static void incr_count(std::pair<int,int>& p)
   {
      ++p.first; ++p.second;
   }
   static int first_index(std::pair<int,int>& p)
   {
      return p.second++;
   }
   static int second_index(std::pair<int,int>& p)
   {
      return (p.second++)-p.first;
   }

public:
   template <typename Graph1, typename Colors1, typename Graph2, typename Colors2>
   static bool prepare_colored(BlissGraph& NG1, const GenericGraph<Graph1>& G1, const Colors1& colors1,
                               BlissGraph& NG2, const GenericGraph<Graph2>& G2, const Colors2& colors2);
   template <typename Graph, typename Colors>
   static bool prepare_colored(BlissGraph& NG, const GenericGraph<Graph>& G, const Colors& colors);

   Array<unsigned int> find_permutation(const BlissGraph& g2) const;
   std::pair< Array<unsigned int>, Array<unsigned int> > find_permutations(const BlissGraph& g2, int n_cols) const;

   int n_automorphisms() const { return n_autom; }
   const std::list< Array<unsigned int> >& automorphisms() const { return autom; }
};

template <typename Graph1, typename Graph2> inline
bool isomorphic(const GenericGraph<Graph1>& G1, const GenericGraph<Graph2>& G2)
{
   if (G1.is_directed != G2.is_directed || G1.nodes() != G2.nodes())
      return false;
   if (G1.nodes()<=1) return true;
   BlissGraph NG1(G1), NG2(G2);
   return NG1==NG2;
}

template <typename Graph1, typename Colors1, typename Graph2, typename Colors2> inline
bool isomorphic(const GenericGraph<Graph1>& G1, const Colors1& colors1,
                const GenericGraph<Graph2>& G2, const Colors2& colors2)
{
   if (G1.is_directed != G2.is_directed || G1.nodes() != G2.nodes())
      return false;
   if (G1.nodes()<=1) return true;
   BlissGraph NG1, NG2;
   return BlissGraph::prepare_colored(NG1, G1, colors1, NG2, G2, colors2) && NG1==NG2;
}

template <typename Graph1, typename Graph2> inline
Array<unsigned int> find_node_permutation(const GenericGraph<Graph1>& G1, const GenericGraph<Graph2>& G2)
{
   if (G1.is_directed != G2.is_directed)
      throw no_match("graphs of different kind");
   if (G1.nodes() != G2.nodes())
      throw no_match("graphs of different size");
   BlissGraph NG1(G1), NG2(G2);
   return NG1.find_permutation(NG2);
}

template <typename Graph1, typename Colors1, typename Graph2, typename Colors2> inline
Array<unsigned int> find_node_permutation(const GenericGraph<Graph1>& G1, const Colors1& colors1,
                                 const GenericGraph<Graph2>& G2, const Colors2& colors2)
{
   if (G1.is_directed != G2.is_directed)
      throw no_match("graphs of different kind");
   if (G1.nodes() != G2.nodes())
      throw no_match("graphs of different size");
   BlissGraph NG1, NG2;
   if (BlissGraph::prepare_colored(NG1, G1, colors1, NG2, G2, colors2))
      return NG1.find_permutation(NG2);
   else
      throw no_match("different colors");
}

template <typename Graph> inline
Array< Array<unsigned int> > automorphisms(const GenericGraph<Graph>& G)
{
   BlissGraph NG(G, true);
   return Array< Array<unsigned int> >(NG.n_automorphisms(), NG.automorphisms().begin());
}

template <typename Graph, typename Colors> inline
Array< Array<unsigned int> > automorphisms(const GenericGraph<Graph>& G, const Colors& colors)
{
   BlissGraph NG;
   BlissGraph::prepare_colored(NG,G,colors);
   return Array< Array<unsigned int> >(NG.n_automorphisms(), NG.automorphisms().begin());
}

template <typename Matrix1, typename Matrix2> inline
typename pm::enable_if<bool, Matrix1::is_symmetric==Matrix2::is_symmetric>::type
isomorphic(const GenericIncidenceMatrix<Matrix1>& M1, const GenericIncidenceMatrix<Matrix2>& M2)
{
   if (M1.rows() != M2.rows() || M1.cols() != M2.cols())
      return false;
   if (M1.rows()<=1) return true;
   BlissGraph NG1(M1), NG2(M2);
   return NG1==NG2;
}

template <typename Matrix1, typename Matrix2> inline
typename pm::enable_if< Array<unsigned int>, Matrix1::is_symmetric && Matrix2::is_symmetric>::type
find_row_permutation(const GenericIncidenceMatrix<Matrix1>& M1, const GenericIncidenceMatrix<Matrix2>& M2)
{
   if (M1.rows() != M2.rows())
      throw no_match("matrices of different dimensions");
   BlissGraph NG1(M1), NG2(M2);
   return NG1.find_permutation(NG2);
}

template <typename Matrix1, typename Matrix2> inline
typename pm::enable_if< std::pair< Array<unsigned int>, Array<unsigned int> >, !Matrix1::is_symmetric && !Matrix2::is_symmetric>::type
find_row_col_permutation(const GenericIncidenceMatrix<Matrix1>& M1, const GenericIncidenceMatrix<Matrix2>& M2)
{
   if (M1.rows() != M2.rows() || M1.cols() != M2.cols())
      throw no_match("matrices of different dimensions");
   BlissGraph NG1(M1), NG2(M2);
   return NG1.find_permutations(NG2, M1.cols());
}

template <typename Matrix> inline
typename pm::enable_if< Array< Array<unsigned int> >, Matrix::is_symmetric>::type
automorphisms(const GenericIncidenceMatrix<Matrix>& M)
{
   BlissGraph NG(M, true);
   return Array< Array<unsigned int> >(NG.n_automorphisms(), NG.automorphisms().begin());
}

template <typename Matrix>
typename pm::disable_if< Array< std::pair< Array<unsigned int>, Array<unsigned int> > >, Matrix::is_symmetric>::type
automorphisms(const GenericIncidenceMatrix<Matrix>& M)
{
   BlissGraph NG(M, true);
   Array< std::pair< Array<unsigned int>, Array<unsigned int> > > result(NG.n_automorphisms());
   std::list< Array<unsigned int> >::const_iterator p=NG.automorphisms().begin();
   const int n_rows=M.rows(), n_cols=M.cols();
   sequence rows(n_cols, n_rows), cols(0, n_cols);

   for (Entire< Array< std::pair< Array<unsigned int>, Array<unsigned int> > > >::iterator r=entire(result); !r.at_end(); ++r, ++p) {
      r->first.append(n_rows, translate(select(*p,rows), -n_cols).begin());
      r->second.append(n_cols, select(*p,cols).begin());
   }
   return result;
}

template <typename Graph1, typename Colors1, typename Graph2, typename Colors2>
bool BlissGraph::prepare_colored(BlissGraph& NG1, const GenericGraph<Graph1>& G1, const Colors1& colors1,
                                 BlissGraph& NG2, const GenericGraph<Graph2>& G2, const Colors2& colors2)
{
   const int n=G1.nodes();
   NG1.p_impl=alloc_impl(n, G1.is_directed);
   NG2.p_impl=alloc_impl(n, G2.is_directed);

   typedef Map<typename pm::identical<typename Colors1::value_type, typename Colors2::value_type>::type,
      unsigned int > color_map_type;
   color_map_type color_map;
   unsigned int color_num = 0U;

   for (typename Entire<Colors1>::const_iterator c=entire(colors1); !c.at_end(); ++c)
      if (!color_map.exists(*c))
         color_map[*c]=color_num++;
   for (typename Entire<Colors2>::const_iterator c=entire(colors2); !c.at_end(); ++c)
      if (!color_map.exists(*c))
         color_map[*c]=color_num++;

   for (typename pm::ensure_features<Colors1, pm::cons<pm::end_sensitive, pm::indexed> >::const_iterator
           c=ensure(colors1, (pm::cons<pm::end_sensitive, pm::indexed>*)0).begin(); !c.at_end(); ++c)
      NG1.color(c.index(), color_map[*c]);
   for (typename pm::ensure_features<Colors2, pm::cons<pm::end_sensitive, pm::indexed> >::const_iterator
           c=ensure(colors2, (pm::cons<pm::end_sensitive, pm::indexed>*)0).begin(); !c.at_end(); ++c) 
      NG2.color(c.index(), color_map[*c]);

   NG1.fill(G1);  NG1.finalize(false);
   NG2.fill(G2);  NG2.finalize(false);
   return true;
}

template <typename Graph, typename Colors>
bool BlissGraph::prepare_colored(BlissGraph& NG, const GenericGraph<Graph>& G, const Colors& colors)
{
   const int n=G.nodes();
   NG.p_impl=alloc_impl(n, G.is_directed);

   typedef Map<typename Colors::value_type, unsigned int > color_map_type;
   color_map_type color_map;
   unsigned int color_num = 0U;

   for (typename Entire<Colors>::const_iterator c=entire(colors); !c.at_end(); ++c)
      if (!color_map.exists(*c))
         color_map[*c]=color_num++;

   for (typename pm::ensure_features<Colors, pm::cons<pm::end_sensitive, pm::indexed> >::const_iterator
           c=ensure(colors, (pm::cons<pm::end_sensitive, pm::indexed>*)0).begin(); !c.at_end(); ++c)
      NG.color(c.index(), color_map[*c]);

   NG.fill(G);  NG.finalize(true);
   return true;
}


} }

#endif // POLYMAKE_GRAPH_COMPARE_H

// Local Variables:
// mode:C++
// c-basic-offset:3
// indent-tabs-mode:nil
// End:
