/* Copyright (c) 1997-2012
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Darmstadt, Germany)
   http://www.polymake.de

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
--------------------------------------------------------------------------------
   $Project: polymake $$Id: inlines_int.h 10617 2012-02-29 10:10:30Z paffenho $
*/

#ifndef POLYMAKE_INTERNAL_INLINES_INT_H
#define POLYMAKE_INTERNAL_INLINES_INT_H

#include <cstdlib>

namespace pm {

inline
long gcd(long a, long b)
{
   if (a<0) a=-a;
   if (b<0) b=-b;
   if (a==0) return b;
   if (b==0) return a;
   if (a==1 || b==1) return 1;

   int k=0;
   while (!((a|b)&1)) {
      a>>=1;  b>>=1;  k++;
   }

   if (a&1)
      while (!(b&1)) b>>=1;
   else
      while (!(a&1)) a>>=1;

   long d=a-b;
   while (d) {
      while (!(d&1)) d>>=1;
      if (d>0)
         a=d;
      else
         b=-d;
      d=a-b;
   }

   return a<<k;
}

inline
long lcm(long a, long b)
{
   return (a/gcd(a,b))*b;
}

/// g=gcd(a,b) && g==p*a+q*b, a==m*g, b==n*g
inline
void gcd_ext(long a, long b, long& g, long& p, long& q, long& m, long& n)
{
   if (a==0) {
      g=b;
      p=q=n=1;
      m=0;
      return;
   }
   if (b==0) {
      g=a;
      p=q=m=1;
      n=0;
      return;
   }

   // We are looking for a 2x2 matrix U s.t. U*(a,b)=(0,g) or (g,0)
   // We start with a unit matrix.
   const bool sw=a<b;
   if (sw) std::swap(a,b);

   long U[2][2]={ { (a<0 ? (a=-a, -1L) : 1L), 0 },
                  { 0, (b<0 ? (b=-b, -1L) : 1L) } };
   long k;

   for (;;) {
      k = a/b;
      // multiply U from left with { {1, -k}, {0, 1} }
      U[0][0] -= k * U[1][0];
      U[0][1] -= k * U[1][1];
      if ((a -= k*b) == 0) {
         g=b;
         n=U[0][0]; m=-U[0][1]; 
         p=U[1][0];  q=U[1][1];
         break;
      }

      k = b/a;
      // multiply U from left with { {1, 0}, {-k, 1} }
      U[1][0] -= k * U[0][0];
      U[1][1] -= k * U[0][1];
      if ((b -= k*a) == 0) {
         g=a;
         p=U[0][0];  q=U[0][1];
         n=-U[1][0]; m=U[1][1];
         break;
      }
   }

   if (sw) {
      std::swap(p,q);
      std::swap(m,n);
   }
}

inline
void div(long a, long b, long& quot, long& rem)
{
   quot=a/b;
   rem=a%b;
}

inline
long div_exact(long a, long b)
{
   return a/b;
}

inline int log2_round(unsigned long x, int round)
{
   if (x<=1UL) return 0;

   int step=(sizeof(unsigned long)*8)/2, log2=step;
   unsigned long pow2=1UL<<step;

   while (x != pow2) {
      if (x < pow2) {
         if (!(step >>= 1)) return log2-1+round;
         pow2 >>= step;
         log2 -= step;
      } else {
         if (!(step >>= 1)) return log2+round;
         pow2 <<= step;
         log2 += step;
      }
   }
   return log2;
}

inline int log2_floor(unsigned long x) { return log2_round(x,0); }
inline int log2_ceil(unsigned long x) { return log2_round(x,1); }

inline bool abs_equal(int x, int y) { return x==y || -x==y; }
inline bool abs_equal(long x, long y) { return x==y || -x==y; }
inline bool abs_equal(long x, int y) { return x==y || -x==y; }
inline bool abs_equal(int x, long y) { return x==y || -x==y; }
inline bool abs_equal(double x, double y) { return x==y || x==-y; }
inline bool abs_equal(double x, long y) { return x==y || x==-y; }
inline bool abs_equal(long x, double y) { return x==y || -x==y; }
inline bool abs_equal(double x, int y) { return x==y || x==-y; }
inline bool abs_equal(int x, double y) { return x==y || -x==y; }

template <typename T> T& inv_sign(T&);
inline int& inv_sign(int& x) { x=-x; return x; }
inline long& inv_sign(long& x) { x=-x; return x; }
inline float& inv_sign(float& x) { x=-x; return x; }
inline double& inv_sign(double& x) { x=-x; return x; }

}

namespace polymake {
   using pm::gcd;
   using pm::gcd_ext;
   using pm::lcm;
   using pm::div_exact;
   using pm::log2_floor;
   using pm::log2_ceil;
   using pm::abs_equal;
   using pm::inv_sign;
}

#endif // POLYMAKE_INTERNAL_INLINES_INT_H

// Local Variables:
// mode:C++
// c-basic-offset:3
// indent-tabs-mode:nil
// End:
