/* Copyright (c) 1997-2012
   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Darmstadt, Germany)
   http://www.polymake.org

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
--------------------------------------------------------------------------------
   $Project: polymake $$Id: RationalFunction.h 10570 2012-02-03 13:13:07Z gawrilow $
*/

#ifndef POLYMAKE_RATIONAL_FUNCTION_H
#define POLYMAKE_RATIONAL_FUNCTION_H

#include "polymake/Polynomial.h"

namespace pm {

template <typename T, typename Coefficient, typename Exponent>
struct is_unipolynomial_type {
   static const bool value=derived_from<T, UniMonomial<Coefficient, Exponent> >::value ||
                           derived_from<T, UniTerm<Coefficient, Exponent> >::value ||
                           derived_from<T, UniPolynomial<Coefficient, Exponent> >::value;
};

template <typename Coefficient, typename Exponent>
class RationalFunction {
public:
   typedef UniPolynomial<Coefficient, Exponent> polynomial_type;
   typedef typename polynomial_type::term_hash term_hash;
   typedef typename polynomial_type::ring_type ring_type;
   typedef typename polynomial_type::monomial_type monomial_type;
   typedef Coefficient coefficient_type;
   typedef Exponent exponent_type;

   /// Construct a zero value.
   RationalFunction() :
      num(),
      den(num.get_ring().one_coef(), num.get_ring()) {}

   /// Construct a zero value of explicitely given ring type.
   explicit RationalFunction(const ring_type& r) :
      num(r),
      den(r.one_coef(), r) {}

   /// Construct a constant value.
   template <typename T>
   explicit RationalFunction(const T& c,
                             typename enable_if<void**, fits_as_coefficient<T, polynomial_type>::value>::type=0) :
      num(c),
      den(num.get_ring().one_coef(), num.get_ring()) {}

   /// Construct a constant value of explicitely given ring type.
   template <typename T>
   RationalFunction(const T& c, const ring_type& r,
                    typename enable_if<void**, fits_as_coefficient<T, polynomial_type>::value>::type=0) :
      num(c, r),
      den(r.one_coef(), r) {}

   /// Construct a value with denominator equal to 1.
   template <typename T>
   explicit RationalFunction(const T& p,
                             typename enable_if<void**, (is_unipolynomial_type<T, Coefficient, Exponent>::value)>::type=0) :
      num(p),
      den(num.get_ring().one_coef(), num.get_ring()) {}

   /// Construct a constant value expressed as ratio of two scalars.
   template <typename T1, typename T2>
   RationalFunction(const T1& c1, const T2& c2,
                    typename enable_if<void**, (fits_as_coefficient<T1, polynomial_type>::value && fits_as_coefficient<T2, polynomial_type>::value)>::type=0) :
      num(c1),
      den(num.get_ring().one_coef(), num.get_ring())
   {
      num /= c2;
   }

   /// Construct a constant value expressed as ratio of two scalars, of an explicitely given ring type.
   template <typename T1, typename T2>
   RationalFunction(const T1& c1, const T2& c2, const ring_type& r,
                    typename enable_if<void**, (fits_as_coefficient<T1, polynomial_type>::value && fits_as_coefficient<T2, polynomial_type>::value)>::type=0) :
      num(c1, r),
      den(r.one_coef(), r)
   {
      num /= c2;
   }

   /// Construct a value with denominator equal to a constant scalar.
   template <typename T1, typename T2>
   RationalFunction(const T1& p1, const T2& c2,
                    typename enable_if<void**, (is_unipolynomial_type<T1, Coefficient, Exponent>::value && fits_as_coefficient<T2, polynomial_type>::value)>::type=0) :
      num(p1),
      den(num.get_ring().one_coef(), num.get_ring())
   {
      num /= c2;
   }

   /// Construct a value with numerator equal to a constant scalar.
   template <typename T1, typename T2>
   RationalFunction(const T1& c1, const T2& p2,
                    typename enable_if<void**, (fits_as_coefficient<T1, polynomial_type>::value && is_unipolynomial_type<T2, Coefficient, Exponent>::value)>::type=0) :
      num(c1, p2.get_ring()),
      den(p2)
   {
      if (is_zero(p2)) throw GMP::ZeroDivide();
      normalize_lc();
   }

   /// Construct a value with given numerator and denominator.
   template <typename T1, typename T2>
   RationalFunction(const T1& p1, const T2& p2,
                    typename enable_if<void**, (is_unipolynomial_type<T1, Coefficient, Exponent>::value && is_unipolynomial_type<T2, Coefficient, Exponent>::value)>::type=0) :
      num(p1.get_ring()),
      den(p1.get_ring())
   {
      if (p1.get_ring() != p2.get_ring()) throw std::runtime_error("RationalFunction - arguments of different rings");
      if (is_zero(p2)) throw GMP::ZeroDivide();
      polynomial_type g, p, q;
      gcd_ext(p1, p2, g, p, q, num, den, false);
      normalize_lc();
   }

#if POLYMAKE_DEBUG
   ~RationalFunction() { POLYMAKE_DEBUG_METHOD(RationalFunction,dump); }
   void dump() const { cerr << *this << std::flush; }
#endif

   const ring_type& get_ring() const { return num.get_ring(); }

   typedef polynomial_type numerator_type;
   typedef polynomial_type denominator_type;

   friend
   const polynomial_type& numerator(const RationalFunction& me) { return me.num; }

   friend
   const polynomial_type& denominator(const RationalFunction& me) { return me.den; }

   void swap(RationalFunction& other)
   {
      num.swap(other.num);
      den.swap(other.den);
   }

   RationalFunction& negate()
   {
      num.negate();
      return *this;
   }

   RationalFunction operator- () const
   {
      return RationalFunction(-num, den, True());
   }

   template <typename T>
   typename enable_if<RationalFunction, (fits_as_coefficient<T, polynomial_type>::value || is_unipolynomial_type<T, Coefficient, Exponent>::value)>::type&
   operator+= (const T& c)
   {
      if (__builtin_expect(!is_zero(c), 1))
         num += den * c;
      return *this;
   }

   RationalFunction& operator+= (const RationalFunction& rf)
   {
      if (__builtin_expect(!rf.num.trivial(), 1)) {
         polynomial_type g, p, q, m, n;
         gcd_ext(den, rf.den, g, p, q, m, n, false);
         m *= rf.num;  m += num * n;
         n *= den;
         gcd_ext(m, n, g, p, q, num, den);
         normalize_lc();
      }
      return *this;
   }

   template <typename T>
   typename enable_if<RationalFunction, (fits_as_coefficient<T, polynomial_type>::value || is_unipolynomial_type<T, Coefficient, Exponent>::value)>::type&
   operator-= (const T& c)
   {
      if (__builtin_expect(!is_zero(c), 1))
         num -= den * c;
      return *this;
   }

   RationalFunction& operator-= (const RationalFunction& rf)
   {
      if (__builtin_expect(!rf.num.trivial(), 1)) {
         polynomial_type g, p, q, m, n;
         gcd_ext(den, rf.den, g, p, q, m, n, false);
         m *= rf.num;  m.negate();  m += num * n;
         n *= den;
         gcd_ext(m, n, g, p, q, num, den);
         normalize_lc();
      }
      return *this;
   }

   template <typename T> friend
   typename enable_if<RationalFunction, (fits_as_coefficient<T, polynomial_type>::value || is_unipolynomial_type<T, Coefficient, Exponent>::value)>::type
   operator+ (const RationalFunction& rf, const T& c)
   {
      if (__builtin_expect(!is_zero(c), 1))
         return RationalFunction(rf.num + rf.den * c, rf.den, True());
      else
         return rf;
   }

   template <typename T> friend
   typename enable_if<RationalFunction, (fits_as_coefficient<T, polynomial_type>::value || is_unipolynomial_type<T, Coefficient, Exponent>::value)>::type
   operator+ (const T& c, const RationalFunction& rf)
   {
      return rf+c;
   }

   template <typename T> friend
   typename enable_if<RationalFunction, (fits_as_coefficient<T, polynomial_type>::value || is_unipolynomial_type<T, Coefficient, Exponent>::value)>::type
   operator- (const RationalFunction& rf, const T& c)
   {
      if (__builtin_expect(!is_zero(c), 1))
         return RationalFunction(rf.num - rf.den * c, rf.den, True());
      else
         return rf;
   }

   template <typename T> friend
   typename enable_if<RationalFunction, (fits_as_coefficient<T, polynomial_type>::value || is_unipolynomial_type<T, Coefficient, Exponent>::value)>::type
   operator- (const T& c, const RationalFunction& rf)
   {
      if (__builtin_expect(!is_zero(c), 1))
         return RationalFunction(rf.den * c - rf.num, rf.den, True());
      else
         return -rf;
   }

   friend
   RationalFunction operator+ (const RationalFunction& rf1, const RationalFunction& rf2)
   {
      if (__builtin_expect(rf1.num.trivial(), 0)) {
         return rf2;
      } else if (__builtin_expect(rf2.num.trivial(), 0)) {
         return rf1;
      } else {
         polynomial_type g, p, q, m, n;
         gcd_ext(rf1.den, rf2.den, g, p, q, m, n, false);
         return RationalFunction(rf1.num * n + rf2.num * m, rf1.den * n);
      }
   }

   friend
   RationalFunction operator- (const RationalFunction& rf1, const RationalFunction& rf2)
   {
      if (__builtin_expect(rf1.num.trivial(), 0)) {
         return -rf2;
      } else if (__builtin_expect(rf2.num.trivial(), 0)) {
         return rf1;
      } else {
         polynomial_type g, p, q, m, n;
         gcd_ext(rf1.den, rf2.den, g, p, q, m, n, false);
         return RationalFunction(rf1.num * n - rf2.num * m, rf1.den * n);
      }
   }

   template <typename T>
   typename enable_if<RationalFunction, fits_as_coefficient<T, polynomial_type>::value>::type&
   operator*= (const T& c)
   {
      num *= c;
      return *this;
   }

   template <typename T>
   typename enable_if<RationalFunction, fits_as_coefficient<T, polynomial_type>::value>::type&
   operator/= (const T& c)
   {
      num /= c;
      return *this;
   }

   template <typename T> friend
   typename enable_if<RationalFunction, fits_as_coefficient<T, polynomial_type>::value>::type
   operator* (const RationalFunction& rf, const T& c)
   {
      if (__builtin_expect(!is_zero(c), 1))
         return RationalFunction(rf.num * c, rf.den, True());
      else
         return RationalFunction(rf.get_ring());
   }

   template <typename T> friend
   typename enable_if<RationalFunction, fits_as_coefficient<T, polynomial_type>::value>::type
   operator* (const T& c, const RationalFunction& rf)
   {
      return rf * c;
   }

   template <typename T> friend
   typename enable_if<RationalFunction, fits_as_coefficient<T, polynomial_type>::value>::type
   operator/ (const RationalFunction& rf, const T& c)
   {
      return RationalFunction(rf.num / c, rf.den, True());
   }

   template <typename T> friend
   typename enable_if<RationalFunction, fits_as_coefficient<T, polynomial_type>::value>::type
   operator/ (const T& c, const RationalFunction& rf)
   {
      if (__builtin_expect(rf.num.trivial(), 0)) {
         throw GMP::ZeroDivide();
      } else if (__builtin_expect(is_zero(c), 0)) {
         return RationalFunction(rf.get_ring());
      } else {
         return RationalFunction(rf.den * c, rf.num, False());
      }
   }

   template <typename T> friend
   typename enable_if<RationalFunction, is_unipolynomial_type<T, Coefficient, Exponent>::value>::type
   operator* (const RationalFunction& rf, const T& p)
   {
      if (__builtin_expect(is_zero(p), 1)) {
         return RationalFunction(rf.get_ring());
      } else if (__builtin_expect(rf.num.trivial(), 0)) {
         return rf;
      } else {
         polynomial_type g, p1, p2, m, n;
         gcd_ext(rf.den, p, g, p1, p2, m, n, false);
         return RationalFunction(rf.num * n, m, True());
      }
   }

   template <typename T> friend
   typename enable_if<RationalFunction, is_unipolynomial_type<T, Coefficient, Exponent>::value>::type
   operator* (const T& p, const RationalFunction& rf)
   {
      return rf * p;
   }

   friend
   RationalFunction operator* (const RationalFunction& rf1, const RationalFunction& rf2)
   {
      if (__builtin_expect(rf1.num.trivial(), 0)) {
         return rf1;
      } else if (__builtin_expect(rf2.num.trivial(), 0)) {
         return rf2;
      } else {
         polynomial_type g, p1, q1, m1, n1, h, p2, q2, m2, n2;
         gcd_ext(rf1.num, rf2.den, g, p1, q1, m1, n1, false);
         gcd_ext(rf1.den, rf2.num, h, p2, q2, m2, n2, false);
         return RationalFunction(m1 * n2, m2 * n1, False());
      }
   }

   template <typename T> friend
   typename enable_if<RationalFunction, is_unipolynomial_type<T, Coefficient, Exponent>::value>::type
   operator/ (const RationalFunction& rf, const T& p)
   {
      if (__builtin_expect(is_zero(p), 1)) {
         throw GMP::ZeroDivide();
      } else if (__builtin_expect(rf.num.trivial(), 0)) {
         return rf;
      } else {
         polynomial_type g, p1, p2, m, n;
         gcd_ext(rf.num, p, g, p1, p2, m, n, false);
         return RationalFunction(m, rf.den * n, False());
      }
   }

   template <typename T> friend
   typename enable_if<RationalFunction, is_unipolynomial_type<T, Coefficient, Exponent>::value>::type
   operator/ (const T& p, const RationalFunction& rf)
   {
      if (__builtin_expect(rf.num.trivial(), 1)) {
         throw GMP::ZeroDivide();
      } else if (__builtin_expect(is_zero(p), 0)) {
         return RationalFunction(rf.get_ring());
      } else {
         polynomial_type g, p1, p2, m, n;
         gcd_ext(rf.num, p, g, p1, p2, m, n, false);
         return RationalFunction(rf.den * n, m, False());
      }
   }

   friend
   RationalFunction operator/ (const RationalFunction& rf1, const RationalFunction& rf2)
   {
      if (__builtin_expect(rf2.num.trivial(), 0)) {
         throw GMP::ZeroDivide();
      } else if (__builtin_expect(rf1.num.trivial(), 0)) {
         return rf1;
      } else {
         polynomial_type g, p1, q1, m1, n1, h, p2, q2, m2, n2;
         gcd_ext(rf1.num, rf2.num, g, p1, q1, m1, n1, false);
         gcd_ext(rf1.den, rf2.den, h, p2, q2, m2, n2, false);
         return RationalFunction(m1 * n2, m2 * n1, False());
      }
   }

   template <typename T>
   typename enable_if<RationalFunction, is_unipolynomial_type<T, Coefficient, Exponent>::value>::type&
   operator*= (const T& p)
   {
      *this = (*this) * p;
      return *this;
   }

   template <typename T>
   typename enable_if<RationalFunction, is_unipolynomial_type<T, Coefficient, Exponent>::value>::type&
   operator/= (const T& p)
   {
      *this = (*this) / p;
      return *this;
   }

   RationalFunction& operator*= (const RationalFunction& rf)
   {
      *this = (*this) * rf;
      return *this;
   }

   RationalFunction& operator/= (const RationalFunction& rf)
   {
      *this = (*this) / rf;
      return *this;
   }

   friend
   bool operator== (const RationalFunction& rf1, const RationalFunction& rf2)
   {
      return rf1.num == rf2.num && rf1.den == rf2.den;
   }

   template <typename T> friend
   typename enable_if<bool, (fits_as_coefficient<T, polynomial_type>::value || is_unipolynomial_type<T, Coefficient, Exponent>::value)>::type
   operator== (const RationalFunction& rf, const T& c)
   {
      return rf.den.unit() && rf.num == c;
   }

   template <typename T> friend
   typename enable_if<bool, (fits_as_coefficient<T, polynomial_type>::value || is_unipolynomial_type<T, Coefficient, Exponent>::value)>::type
   operator== (const T& c, const RationalFunction& rf)
   {
      return rf == c;
   }

   friend
   bool operator!= (const RationalFunction& rf1, const RationalFunction& rf2)
   {
      return !(rf1 == rf2);
   }

   template <typename T> friend
   typename enable_if<bool, (fits_as_coefficient<T, polynomial_type>::value || is_unipolynomial_type<T, Coefficient, Exponent>::value)>::type
   operator!= (const RationalFunction& rf, const T& c)
   {
      return !(rf == c);
   }

   template <typename T> friend
   typename enable_if<bool, (fits_as_coefficient<T, polynomial_type>::value || is_unipolynomial_type<T, Coefficient, Exponent>::value)>::type
   operator!= (const T& c, const RationalFunction& rf)
   {
      return !(rf == c);
   }

   template <typename Output> friend
   Output& operator<< (GenericOutput<Output>& out, const RationalFunction& rf)
   {
      out.top() << '(' << rf.num << ")/(" << rf.den << ')';
      return out.top();
   }

protected:
   polynomial_type num, den;

   /// internally used constructor, when the operands are known to be mutually prime and normalized
   template <bool lc_normalized>
   RationalFunction(const polynomial_type& num_arg, const polynomial_type& den_arg, bool2type<lc_normalized>) :
      num(num_arg),
      den(den_arg)
   {
      if (!lc_normalized) normalize_lc();
   }

   void normalize_lc()
   {
      if (num.trivial()) {
         den=polynomial_type(get_ring().one_coef(), get_ring());
      } else {
         const Coefficient den_lc=den.lc();
         if (!is_one(den_lc)) {
            num /= den_lc;
            den /= den_lc;
         }
      }
   }

   template <typename> friend struct spec_object_traits;
};

template <typename Coefficient, typename Exponent> inline
RationalFunction<Coefficient, Exponent>& inv_sign(RationalFunction<Coefficient, Exponent>& rf) { rf.negate(); return rf; }

template <typename Coefficient, typename Exponent>
struct spec_object_traits< Serialized< RationalFunction<Coefficient, Exponent> > > :
   spec_object_traits<is_composite> {

   typedef RationalFunction<Coefficient, Exponent> masquerade_for;

   typedef cons<typename RationalFunction<Coefficient, Exponent>::term_hash,
           cons<typename RationalFunction<Coefficient, Exponent>::term_hash,
                typename RationalFunction<Coefficient, Exponent>::ring_type> > elements;

   template <typename Me, typename Visitor>
   static void visit_elements(Me& me, Visitor& v)
   {
      v << me.num.data->the_terms << me.den.data->the_terms << me.num.data->ring;
      set_den_ring(me.num, me.den);
   }

private:
   static void set_den_ring(const UniPolynomial<Coefficient, Exponent>&, const UniPolynomial<Coefficient, Exponent>&) {}

   static void set_den_ring(UniPolynomial<Coefficient, Exponent>& num, UniPolynomial<Coefficient, Exponent>& den)
   {
      den.data->ring=num.data->ring;
   }
};

template <typename Coefficient, typename Exponent>
struct choose_generic_object_traits< RationalFunction<Coefficient, Exponent>, false, false > :
      spec_object_traits< RationalFunction<Coefficient, Exponent> > {
   typedef void generic_type;
   typedef is_polynomial generic_tag;
   typedef RationalFunction<Coefficient, Exponent> persistent_type;

   static
   bool is_zero(const persistent_type& p)
   {
      return numerator(p).trivial();
   }

   static
   bool is_one(const persistent_type& p)
   {
      return numerator(p).unit() && denominator(p).unit();
   }

   static
   const persistent_type& zero()
   {
      static const persistent_type x=persistent_type();
      return x;
   }

   static
   const persistent_type& one()
   {
      static const persistent_type x(1);
      return x;
   }
};

template <typename Coefficient, typename Exponent>
struct algebraic_traits< UniPolynomial<Coefficient, Exponent> > {
   typedef RationalFunction<typename algebraic_traits<Coefficient>::field_type, Exponent> field_type;
};

template <typename Coefficient, typename Exponent>
struct algebraic_traits< RationalFunction<Coefficient, Exponent> > {
   typedef RationalFunction<typename algebraic_traits<Coefficient>::field_type, Exponent> field_type;
};

template <typename Coefficient, typename Exponent, typename T, typename TModel>
struct isomorphic_types<RationalFunction<Coefficient, Exponent>, T,
                        typename enable_if<is_polynomial, (fits_as_coefficient<T, UniPolynomial<Coefficient, Exponent> >::value ||
                                                           is_unipolynomial_type<T, Coefficient, Exponent>::value)>::type,
                        TModel> : False {
   typedef cons<is_polynomial, is_scalar> discriminant;
};

template <typename Coefficient, typename Exponent, typename T, typename TModel>
struct isomorphic_types<T, RationalFunction<Coefficient, Exponent>, TModel,
                        typename enable_if<is_polynomial, (fits_as_coefficient<T, UniPolynomial<Coefficient, Exponent> >::value ||
                                                           is_unipolynomial_type<T, Coefficient, Exponent>::value)>::type> : False {
   typedef cons<is_scalar, is_polynomial> discriminant;
};

template <typename Coefficient, typename Exponent>
struct isomorphic_types<RationalFunction<Coefficient, Exponent>, RationalFunction<Coefficient, Exponent>, is_polynomial, is_polynomial> : True {
   typedef cons<is_polynomial, is_polynomial> discriminant;
};


template <typename Coefficient, typename Exponent, typename T> inline
typename enable_if<RationalFunction<Coefficient, Exponent>,
                   is_unipolynomial_type<T, Coefficient, Exponent>::value>::type
operator/ (const UniPolynomial<Coefficient, Exponent>& p1, const T& p2)
{
   return RationalFunction<Coefficient, Exponent>(p1, p2);
}

template <typename Coefficient, typename Exponent, typename T> inline
typename enable_if<RationalFunction<Coefficient, Exponent>,
                   is_unipolynomial_type<T, Coefficient, Exponent>::value>::type
operator/ (const UniTerm<Coefficient, Exponent>& p1, const T& p2)
{
   return RationalFunction<Coefficient, Exponent>(p1, p2);
}

template <typename Coefficient, typename Exponent, typename T> inline
typename enable_if<RationalFunction<Coefficient, Exponent>,
                   is_unipolynomial_type<T, Coefficient, Exponent>::value>::type
operator/ (const UniMonomial<Coefficient, Exponent>& p1, const T& p2)
{
   return RationalFunction<Coefficient, Exponent>(p1, p2);
}

namespace operations {

// operations neg, add, sub, mul defined in Polynomial.h will instantiate correctly for RationalFunction too.

template <typename LeftRef, class RightRef>
struct div_impl<LeftRef, RightRef,
                typename enable_if< cons<is_polynomial, is_polynomial>, is_instance2_of<typename deref<LeftRef>::type, RationalFunction>::value>::type> {
   typedef LeftRef first_argument_type;
   typedef RightRef second_argument_type;
   typedef typename deref<LeftRef>::type result_type;

   result_type operator() (typename function_argument<LeftRef>::const_type l,
                           typename function_argument<RightRef>::const_type r) const
   {
      return l/r;
   }

   void assign(typename lvalue_arg<LeftRef>::type l, typename function_argument<RightRef>::const_type r) const
   {
      l/=r;
   }
};

template <typename LeftRef, class RightRef>
struct div_impl<LeftRef, RightRef,
                typename enable_if< cons<is_scalar, is_polynomial>, is_instance2_of<typename deref<RightRef>::type, RationalFunction>::value>::type> {
   typedef LeftRef first_argument_type;
   typedef RightRef second_argument_type;
   typedef typename deref<RightRef>::type result_type;

   result_type operator() (typename function_argument<LeftRef>::const_type l,
                           typename function_argument<RightRef>::const_type r) const
   {
      return l/r;
   }
};

template <typename LeftRef, class RightRef>
struct div_impl<LeftRef, RightRef,
                typename enable_if< cons<is_polynomial, is_scalar>, is_instance2_of<typename deref<LeftRef>::type, RationalFunction>::value>::type> {
   typedef LeftRef first_argument_type;
   typedef RightRef second_argument_type;
   typedef typename deref<LeftRef>::type result_type;

   result_type operator() (typename function_argument<LeftRef>::const_type l,
                           typename function_argument<RightRef>::const_type r) const
   {
      return l/r;
   }

   void assign(typename lvalue_arg<LeftRef>::type l, typename function_argument<RightRef>::const_type r) const
   {
      l/=r;
   }
};

} }

namespace polymake {
   using pm::RationalFunction;
}

namespace std {
   template <typename Coefficient, typename Exponent>
   void swap(pm::RationalFunction<Coefficient,Exponent>& x1, pm::RationalFunction<Coefficient,Exponent>& x2) { x1.swap(x2); }
}

#endif // POLYMAKE_RATIONAL_FUNCTION_H

// Local Variables:
// mode:C++
// c-basis-offset:3
// indent-tabs-mode:nil
// End:
